/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.listener;

import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.transaction.NotSupportedException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;

import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.Message;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.jbidescriptor.generated.ExchangeProperties;
import org.ow2.petals.component.framework.jbidescriptor.generated.MEPType;
import org.ow2.petals.component.framework.jbidescriptor.generated.MessageParam;
import org.ow2.petals.component.framework.jbidescriptor.generated.MessageProperties;
import org.ow2.petals.component.framework.jbidescriptor.generated.Param;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.message.ExchangeImpl;
import org.ow2.petals.component.framework.notification.notify.FaultProducerComponentNotify;
import org.ow2.petals.component.framework.notification.notify.OutProducerComponentNotify;
import org.ow2.petals.component.framework.notification.notify.StatusProducerComponentNotify;
import org.ow2.petals.component.framework.process.MessageInterceptorProcessor;
import org.ow2.petals.component.framework.process.async.AsyncContext;
import org.ow2.petals.component.framework.process.async.AsyncMessageManager;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.util.UtilFactory;

/**
 * The abstract methods for a CDK listener. Both external and JBI listener
 * extend it.
 * 
 * @author chamerling - EBM WebSourcing
 * @author ofabre - EBM WebSourcing
 * @author alouis - EBM WebSourcing
 * 
 */
public abstract class AbstractListener {

    /**
     * The JBI component itself
     */
    protected AbstractComponent component;

    /**
     * The messageExchange factory used to create MessageExchanges
     */
    protected MessageExchangeFactory factory;

    /**
     * The consumes for External Listeners
     */
    private Consumes consumes;

    /**
     * The provides for JBI listeners
     */
    private Provides provides;

    /**
     * The Consumes or Provides extensions
     */
    private ConfigurationExtensions extensions;

    /**
     * The current Transaction
     */
    private Transaction currentTransaction;

    /**
     * The message interceptor processor; Each listener have its own interceptor
     * processor in order to get better performances.
     */
    protected MessageInterceptorProcessor messageInterceptorProcessor;

    /**
     * Creates a new instance of {@link AbstractListener}
     * 
     */
    public AbstractListener() {
        super();
    }

    /**
     * Create a message exchange with all the information set in the specified
     * CONSUME <br>
     * - address (interface, service, endpoint) <br>
     * - MEP <br>
     * - operation <br>
     * - PEtALS exchange' meta-data (org.objectweb.petals.xxx)
     * 
     * @return a message exchange set with Consumes attributes
     * @throws MessagingException
     *             The message exchange can't be created or no pattern specified
     *             in extensions
     * @throws PEtALSCDKException
     *             The current Transaction has failed to return its status
     */
    public Exchange createConsumeExchange(final Consumes consumes) throws MessagingException,
    PEtALSCDKException {
        final MEPConstants mep = this.retrievePattern(consumes);
        final Exchange exchange = this.createConsumeExchange(consumes, mep);
        return exchange;
    }

    /**
     * Create a message exchange with all the information set in the specified
     * CONSUME, with the given exchange pattern <br>
     * - address (interface, service, endpoint) <br>
     * - operation <br>
     * - PEtALS exchange' meta-data (org.objectweb.petals.xxx)
     * 
     * @return a message exchange set with Consumes attributes
     * @throws MessagingException
     *             The message exchange can't be created or no pattern specified
     *             in extensions
     * @throws PEtALSCDKException
     *             The current Transaction has failed to return its status
     */
    public Exchange createConsumeExchange(final Consumes consumes, final MEPConstants mep)
    throws MessagingException, PEtALSCDKException {
        final Exchange exchange = this.createExchange(mep);
        this.setOperation(exchange, consumes);
        this.setProperties(exchange, consumes);
        this.setEndpoint(exchange, consumes);
        return exchange;
    }

    /**
     * Create an exchange implementation matching the given pattern.
     * 
     * @param mep
     *            a specific MEP pattern
     * @return an exchange implementation matching the given pattern
     * @throws MessagingException
     *             The message exchange can't be created or no pattern specified
     *             in extensions
     * @throws PEtALSCDKException
     *             The current Transaction has failed to return its status
     */
    public Exchange createExchange(final MEPConstants mep) throws MessagingException,
    PEtALSCDKException {
        MessageExchange exchange = null;
        if (mep == null) {
            exchange = this.factory.createExchange(null);
        } else {
            exchange = this.factory.createExchange(mep.value());
        }
        // if there is a current transaction,
        try {
            if (this.currentTransaction != null
                    && this.currentTransaction.getStatus() == Status.STATUS_ACTIVE) {
                exchange.setProperty(MessageExchange.JTA_TRANSACTION_PROPERTY_NAME,
                        this.currentTransaction);
            }
        } catch (final SystemException e) {
            throw new PEtALSCDKException("Failed to get current transaction", e);
        }

        return new ExchangeImpl(exchange);
    }

    /**
     * Get the component.
     * 
     * @return
     */
    public AbstractComponent getComponent() {
        return this.component;
    }

    /**
     * The direct usage of the Component context is not recommended in the CDK.
     * 
     * @return
     */
    @Deprecated
    public ComponentContext getComponentContext() {
        return this.component.getContext();
    }

    public Consumes getConsumes() {
        return this.consumes;
    }

    /**
     * Return the current Transaction.
     * 
     * @return The current Transaction or {@code null} if no Transaction Manager
     *         is implemented by the JBI container
     * 
     * @throws SystemException
     * @throws NotSupportedException
     */
    public Transaction getCurrentTransaction() throws PEtALSCDKException {
        if (this.currentTransaction == null) {
            final TransactionManager transactionManager = (TransactionManager) this.component
            .getContext().getTransactionManager();
            if (transactionManager != null) {
                try {
                    transactionManager.begin();
                    this.currentTransaction = transactionManager.getTransaction();
                } catch (final NotSupportedException e) {
                    throw new PEtALSCDKException("Failed to create a Transaction", e);
                } catch (final SystemException e) {
                    throw new PEtALSCDKException("Failed to create a Transaction", e);
                }

            }
        }
        return this.currentTransaction;
    }

    /**
     * Get the service extensions.
     * 
     * @return Extensions, can be null if no extensions found
     */
    public ConfigurationExtensions getExtensions() {
        return this.extensions;
    }

    public String getInstallRoot() {
        final AbstractServiceUnitManager suManager = (AbstractServiceUnitManager) this
        .getComponent().getServiceUnitManager();
        String installRoot = null;
        if (this.consumes != null) {
            installRoot = suManager.getSUDataHandlerForConsumes(this.consumes).getInstallRoot();
        } else {
            installRoot = suManager.getSUDataHandlerForProvides(this.provides).getInstallRoot();
        }
        return installRoot;
    }

    /**
     * 
     * @return
     */
    public Logger getLogger() {
        return this.component.getLogger();
    }

    /**
     * @return the messageInterceptorProcessor
     */
    public MessageInterceptorProcessor getMessageInterceptorProcessor() {
        return this.messageInterceptorProcessor;
    }

    public Provides getProvides() {
        return this.provides;
    }

    /**
     * A convenience method which can be overridden so that there's no need to
     * call super.init(). This method is called after setting of all Listener
     * attributes
     * 
     */
    public void init() {
    }

    /**
     * Initialize the listener.
     * 
     * @param component
     */
    public final void init(final AbstractComponent component) {
        this.component = component;
        this.messageInterceptorProcessor = new MessageInterceptorProcessor(this.component, this
                .getLogger());
        this.factory = this.getComponent().getChannel().createExchangeFactory();
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * asynchronous way.<br/>
     * Process MessageExchange interceptors before sending message. <br/>
     * Process the notifications before sending message No Asynchronous context
     * management is proposed.
     * 
     * @param exchange
     *            message exchange to send; must be non-null
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public void send(Exchange exchange) throws MessagingException {
        this.component.getLogger().fine(
                "Send an asynchronous exchange with id : " + exchange.getExchangeId());

        if (exchange.getProperty(AsyncMessageManager.ASYNC_ORIGINAL_MESSAGE_PROPERTY) != null) {
            this.component.getAsyncManager().removeAsyncContext(exchange.getExchangeId());
            try {
                this.interceptAcceptResponse(exchange);
                exchange = this.processResponseNotification(exchange);
            } catch (final PEtALSCDKException e) {
                throw new MessagingException(e);
            }
        } else {
            try {
                this.interceptSend(exchange);
            } catch (final PEtALSCDKException e) {
                throw new MessagingException(e);
            }
        }
        this.getComponent().getChannel().send(((ExchangeImpl) exchange).getMessageExchange());
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * asynchronous way.<br/>
     * 
     * @param exchange
     *            message exchange to send; must be non-null
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public void sendAsync(final Exchange exchange, final AsyncContext asyncContext)
    throws MessagingException {
        this.component.getAsyncManager().addAsyncContext(exchange, asyncContext);
        try {
            this.send(exchange);
        } catch (final MessagingException e) {
            this.component.getAsyncManager().removeAsyncContext(exchange.getExchangeId());
            throw e;
        }
    }

    /**
     * Call sendSync(MessageExchage exchange, long timeoutMS) with a timeout
     * value found in provides or consumes extensions or "0" if not found
     * 
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public boolean sendSync(final Exchange exchange) throws MessagingException {
        long timeoutMS = Constants.Component.DEFAULT_SEND_TIMEOUT;

        // set the timeout if it is configured
        if (this.consumes != null) {
            if (this.consumes.getTimeout() != null) {
                timeoutMS = Long.valueOf(this.consumes.getTimeout());
            }
        } else if (this.provides != null) {
            if (this.provides.getTimeout() != null) {
                timeoutMS = Long.valueOf(this.provides.getTimeout());
            }
        }

        return this.sendSync(exchange, timeoutMS);
    }

    /**
     * Send a message to the endpoint specified in the MessageExchange in an
     * synchronous way.<br/>
     * Process MessageExchange interceptors before sending message. <br/>
     * Add performance notification
     * 
     * @param exchange
     *            message exchange to send; must be non-null
     * @param timeoutMS
     *            timeout period, in milliseconds (0 means no timeout); must be
     *            greater than or equal to zero
     * @return true if the message has been sent (if a timeout occurs, return
     *         false)
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message through the container
     * @since 2.0 operation and metadata are now set with
     *        createConsumeExchange()
     */
    public boolean sendSync(final Exchange exchange, final long timeoutMS)
    throws MessagingException {
        boolean bRet = false;

        try {
            this.interceptSend(exchange);
            this.component.getLogger().finest(
                    "Send a synchronous exchange with id : " + exchange.getExchangeId());
            bRet = this.getComponent().getChannel().sendSync(
                    ((ExchangeImpl) exchange).getMessageExchange(), timeoutMS);
            this.interceptSendResponse(exchange);
        } catch (final PEtALSCDKException e) {
            throw new MessagingException(e);
        } catch (final JBIException e) {
            throw new MessagingException(e);
        }

        return bRet;
    }

    /**
     * @param consumes
     *            the consumes to set
     */
    public final void setConsumes(final Consumes consumes) {
        this.consumes = consumes;
    }

    /**
     * 
     * @param transaction
     */
    public final void setCurrentTransaction(final Transaction transaction) {
        this.currentTransaction = transaction;
    }

    /**
     * @param consumes
     *            the consumes to set
     */
    public final void setExtensions(final ConfigurationExtensions extensions) {
        this.extensions = extensions;
    }

    /**
     * Set Provides section
     * 
     * @param provides
     *            the provides to set
     */
    public final void setProvides(final Provides provides) {
        this.provides = provides;
    }

    /**
     * Retrieve the MEP from the consumer node extensions.
     * 
     * @param consumes
     *            The consumer node of the SU JBI descriptor.
     * @return MEP specified in extensions, or <code>null</code> if no MEP
     *         pattern specified.
     * @throws MessagingException
     *             MEP pattern specified in consumes extension is not valid.
     */
    protected final MEPConstants retrievePattern(final Consumes consumes) throws MessagingException {
        MEPConstants mep = null;

        if (consumes.getMep() != null) {
            if (MEPType.IN_ONLY.equals(consumes.getMep())) {
                mep = MEPConstants.IN_ONLY_PATTERN;
            } else if (MEPType.ROBUST_IN_ONLY.equals(consumes.getMep())) {
                mep = MEPConstants.ROBUST_IN_ONLY_PATTERN;
            } else if (MEPType.IN_OUT.equals(consumes.getMep())) {
                mep = MEPConstants.IN_OUT_PATTERN;
            } else if (MEPType.IN_OPTIONAL_OUT.equals(consumes.getMep())) {
                mep = MEPConstants.IN_OPTIONAL_OUT_PATTERN;
            } else {
                throw new MessagingException("MEP pattern '" + consumes.getMep().value()
                        + "' is not valid");
            }
        }

        return mep;
    }

    /**
     * Call the Interceptors before sending the request exchange
     * 
     * @param exchange
     * @throws PEtALSCDKException
     */
    private void interceptSend(final Exchange exchange) throws PEtALSCDKException {
        if (this.consumes != null) {
            this.messageInterceptorProcessor.intercept(exchange, this.consumes,
                    Message.InterceptorPhase.SEND);
        }
    }

    /**
     * Call the Interceptors after a synchronous send.
     * 
     * @param exchange
     * @throws PEtALSCDKException
     */
    private void interceptSendResponse(final Exchange exchange) throws PEtALSCDKException {
        if (this.consumes != null) {
            this.messageInterceptorProcessor.intercept(exchange, this.consumes,
                    Message.InterceptorPhase.SEND_RESPONSE);
        }
    }

    /**
     * Call the Intercectors before an async message response
     * 
     * @param exchange
     * @throws PEtALSCDKException
     */
    private void interceptAcceptResponse(final Exchange exchange) throws PEtALSCDKException {
        final Provides provides = ((AbstractServiceUnitManager) this.getComponent()
                .getServiceUnitManager()).getProvidesFromEndpoint(exchange.getEndpoint());
        if (provides != null) {
            this.messageInterceptorProcessor.intercept(exchange, provides,
                    Message.InterceptorPhase.ACCEPT_RESPONSE);
        }
    }

    /**
     * Process the out, fault and status notifications
     * 
     * @param exchange
     *            The exchange to process
     * @return The exchange without any change
     */
    private Exchange processResponseNotification(final Exchange exchange) {
        // TODO implements a great notification manager, intercepter layer on
        // the send activities
        ((ExchangeImpl) exchange).setCheckRole(false);
        if (this.component.getNotificationBrokerController() != null
                && !UtilFactory.getExchangeUtil().isBaseNotification(exchange)) {
            if (exchange.isInOnlyPattern()) {
                this.component.getNotificationBrokerController().getNotificationProducer()
                .getNotifyController().process(StatusProducerComponentNotify.class,
                        exchange);
            } else if (exchange.isRobustInOnlyPattern()) {
                if (exchange.isActiveStatus()) {
                    if (exchange.getFault() == null) {
                        this.component.getNotificationBrokerController().getNotificationProducer()
                        .getNotifyController().process(OutProducerComponentNotify.class,
                                exchange);
                    } else {
                        this.component.getNotificationBrokerController().getNotificationProducer()
                        .getNotifyController().process(FaultProducerComponentNotify.class,
                                exchange);
                    }
                } else {
                    this.component.getNotificationBrokerController().getNotificationProducer()
                    .getNotifyController().process(StatusProducerComponentNotify.class,
                            exchange);
                }
            } else if (exchange.isInOutPattern()) {
                if (exchange.getFault() == null) {
                    this.component.getNotificationBrokerController().getNotificationProducer()
                    .getNotifyController().process(OutProducerComponentNotify.class,
                            exchange);
                } else {
                    this.component.getNotificationBrokerController().getNotificationProducer()
                    .getNotifyController().process(FaultProducerComponentNotify.class,
                            exchange);
                }
            } else if (exchange.isInOptionalOutPattern()) {
                if (exchange.isActiveStatus()) {
                    if (exchange.getFault() == null) {
                        this.component.getNotificationBrokerController().getNotificationProducer()
                        .getNotifyController().process(OutProducerComponentNotify.class,
                                exchange);
                    } else {
                        this.component.getNotificationBrokerController().getNotificationProducer()
                        .getNotifyController().process(FaultProducerComponentNotify.class,
                                exchange);
                    }
                } else {
                    this.component.getNotificationBrokerController().getNotificationProducer()
                    .getNotifyController().process(StatusProducerComponentNotify.class,
                            exchange);
                }
            }
        }
        ((ExchangeImpl) exchange).setCheckRole(true);
        return exchange;
    }

    /**
     * Fill the exchange with the service information (itf,src, and ep) from the
     * consume extension
     * 
     * @param exchange
     */
    private void setEndpoint(final Exchange exchange, final Consumes consumes) {

        // Retrieve possible target endpoint
        ServiceEndpoint ep = null;
        if (consumes.getEndpointName() != null && consumes.getServiceName() != null) {
            ep = this.getComponent().getContext().getEndpoint(consumes.getServiceName(),
                    consumes.getEndpointName());
        }

        // set exchange information
        exchange.setInterfaceName(consumes.getInterfaceName());
        exchange.setService(consumes.getServiceName());
        exchange.setEndpoint(ep);

        if (consumes.getLinkType() != null) {
            exchange.setProperty(Exchange.PROPERTY_PROVIDER_LINKTYPE, consumes.getLinkType()
                    .value());
        }
    }

    /**
     * Set operation to call if exists in extensions. If no extension can be
     * found or if operation value is "" or missing, operation isn't set.
     * 
     * @param exchange
     *            a {@link MessageExchange}, must not be null
     * @param extensions
     */
    private final void setOperation(final Exchange exchange, final Consumes consumes) {

        if (consumes.getOperation() != null) {
            exchange.setOperation(consumes.getOperation());
        }
    }

    /**
     * Set exchange and message properties from extensions.
     * 
     * @param exchange
     *            a {@link Exchange}
     * @param extensions
     */
    private final void setProperties(final Exchange exchange, final Consumes consumes) {
        final ExchangeProperties exchangeProperties = consumes.getExchangeProperties();
        if (exchangeProperties != null) {
            for (final Param exchangeProperty : exchangeProperties.getExchangeProperty()) {
                exchange.setProperty(exchangeProperty.getName(), exchangeProperty.getValue());
            }
        }

        final MessageProperties messageProperties = consumes.getMessageProperties();
        if (messageProperties != null) {
            for (final MessageParam messageProperty : messageProperties.getMessageProperty()) {
                final NormalizedMessage message = ((ExchangeImpl) exchange).getMessageExchange()
                .getMessage(messageProperty.getMessage().name());
                message.setProperty(messageProperty.getName(), messageProperty.getValue());
            }
        }
    }
}
